"""
-Should store moves as array of locations e.g.: [[x1,y1],[x2,y2]]
if self.player_turn == True then it is player 1's turn
"""
from checker_players.Interpretters.DoubleColumnInterpretter import DoubleColumnInterpretter
import copy
class Board:
    """
    A class to represent and play an 8x8 game of checkers.
    """
    EMPTY_SPOT = 0
    P1 = 1
    P2 = 2
    P1_K = 3
    P2_K = 4
    BACKWARDS_PLAYER = P2
    HEIGHT = 8
    WIDTH = 8

    Current_Winner = 0
    game_over = False
    
    def __init__(self, board_state=None, P1_turn=True):
        """
        Initializes a new instance of the Board class.  Unless specified otherwise, 
        the board will be created with a start board configuration.
        """
        self.player_turn = P1_turn 
        self.spots = [ [self.EMPTY_SPOT] * self.WIDTH for _ in range(self.HEIGHT)]
        
        if board_state is None:
            for x in range(self.WIDTH):
                for y in range(self.HEIGHT):   
                    if y<3 and (x+y)%2==1:
                        self.spots[x][y] = self.P1
                    elif y>4 and (x+y)%2==1:
                        self.spots[x][y] = self.P2
        else:
            self.spots = copy.deepcopy(board_state)

    def reset_board(self):
        """
        Resets the current configuration of the game board to the original 
        starting position.
        """
        self.spots = Board().spots
        self.Current_Winner = 0
        self.game_over = False
        
    def empty_board(self):
        """
        Removes any pieces currently on the board and leaves the board with nothing but empty spots.
        """
        self.spots = [ [self.EMPTY_SPOT] * self.WIDTH for _ in range(self.HEIGHT)]
    
    def is_game_over(self):
        """
        Finds out and returns whether the game currently being played is over or
        not.
        """
        player1_pieces = [[x,y] for x in range(self.WIDTH) for y in range(self.HEIGHT) if self.get_spot_info([x,y]) in (self.P1, self.P1_K)]
        player2_pieces = [[x,y] for x in range(self.WIDTH) for y in range(self.HEIGHT) if self.get_spot_info([x,y]) in (self.P2, self.P2_K)]

        possible_moves = self.get_possible_next_moves()

        return (not player1_pieces) or (not player2_pieces) or (not possible_moves)

    def is_spot(self, loc):
        """
        Finds out of the spot at the given location is an actual spot on the game board.
        """
        return loc[0] >= 0 and loc[0] < self.WIDTH and loc[1] >= 0 and loc[1] < self.HEIGHT
      
    def get_spot_info(self, loc):
        """
        Gets the information about the spot at the given location.
        """
        return self.spots[loc[0]][loc[1]]
       
    def forward_n_locations(self, start_loc, n, backwards=False):
        """
        Gets the locations possible for moving a piece from a given location diagonally
        forward (or backwards if wanted) a given number of times(without directional change midway).  
        """
        if backwards:
            return [[start_loc[0]+(x_parity*n), start_loc[1]+(-1*n)] for x_parity in [1,-1]]
        else:
            return [[start_loc[0]+(x_parity*n), start_loc[1]+(n)] for x_parity in [1,-1]]
    
    def is_king(self, loc):
        piece = self.spots[loc[0]][loc[1]]
        return piece == self.P1_K or piece == self.P2_K

    def is_backwards_player(self, loc):
        piece = self.spots[loc[0]][loc[1]]
        return piece == self.BACKWARDS_PLAYER

    def is_empty(self, loc):
        piece = self.spots[loc[0]][loc[1]]
        return piece == self.EMPTY_SPOT
    
    def get_simple_moves(self, start_loc):
        """    
        Gets the possible moves a piece can make given that it does not capture any opponents pieces.
        
        PRE-CONDITION:
        -start_loc is a location with a players piece
        """
        
        next_locations = []
        if self.is_king(start_loc):
            next_locations = self.forward_n_locations(start_loc, 1)
            next_locations.extend(self.forward_n_locations(start_loc, 1, backwards=True))
        elif self.is_backwards_player(start_loc):
            next_locations = self.forward_n_locations(start_loc, 1, backwards=True)
        else:
            next_locations = self.forward_n_locations(start_loc, 1)
        

        possible_next_locations = [loc for loc in next_locations if self.is_spot(loc) and self.is_empty(loc)]
            
        return [[start_loc, end_spot] for end_spot in possible_next_locations]   
       
    def are_opponents(self, loc1, loc2):
        piece1 = self.spots[loc1[0]][loc1[1]]
        piece2 = self.spots[loc2[0]][loc2[1]]
        if (piece1 == self.P1 or piece1 == self.P1_K):
            return piece2 == self.P2 or piece2 == self.P2_K 
        else:
            return piece2 == self.P1 or piece2 == self.P1_K      
 
    def get_capture_moves(self, start_loc):
        """
        Get all of the possible moves for a piece which involve capturing an opponent's piece.
        """
            
        answer = []
        next1 = []
        next2 = []
        
        if self.is_king(start_loc):  
            next1 = self.forward_n_locations(start_loc, 1)
            next2 = self.forward_n_locations(start_loc, 2)
            next1.extend(self.forward_n_locations(start_loc, 1, True))
            next2.extend(self.forward_n_locations(start_loc, 2, True))
        elif self.is_backwards_player(start_loc):
            next1 = self.forward_n_locations(start_loc, 1, True)
            next2 = self.forward_n_locations(start_loc, 2, True)
        else:
            next1 = self.forward_n_locations(start_loc, 1)
            next2 = self.forward_n_locations(start_loc, 2)

        for loc1, loc2 in zip(next1, next2):
            if self.is_spot(loc1) and self.is_spot(loc2) and self.is_empty(loc2) and self.are_opponents(start_loc, loc1):
                capture_move = [start_loc, loc2]
                answer.append(capture_move)
                            
        return answer
        
    def get_possible_next_moves(self):
        """
        Gets the possible moves that can be made from the current board configuration.
        """
        flat_map = lambda xs: [y for ys in xs for y in ys]
        double_column = DoubleColumnInterpretter().interpret(self)
        current_pieces = double_column[self.P1 if self.player_turn else self.P2]
        
        simple_moves = flat_map([self.get_simple_moves(loc) for loc in current_pieces if self.get_simple_moves(loc) != []])
        capture_moves = flat_map([self.get_capture_moves(loc) for loc in current_pieces if self.get_capture_moves(loc) != []])
        return simple_moves + capture_moves 

    def get_kings_row_positions(self, player):
        if player == self.BACKWARDS_PLAYER:
            return [[i, 0] for i in range(self.WIDTH)]
        elif player in (self.P1, self.P2):
            return [[i, self.HEIGHT-1] for i in range(self.WIDTH)]
        else:
            raise ValueError("Not a valid player number. Availible player numbers are: " +str(self.P1)+", "+str(self.P2))

    def make_move(self, move, switch_player_turn=True):
        """
        Makes a given move on the board, and (as long as is wanted) switches the indicator for
        which players turn it is.
        """
        init_loc = move[0]
        final_loc = move[1]
        current_player = self.P1 if self.player_turn else self.P2

        # Here we are just making sure that the right player is moving their piece
        if (self.get_spot_info(init_loc) in (self.P1, self.P1_K) and not self.player_turn) or (self.get_spot_info(init_loc) in (self.P2, self.P2_K) and self.player_turn):
            raise ValueError("Piece of type " + str(self.get_spot_info(init_loc)) + " was attempted to move while Board.switch_player_turn = " + str(self.player_turn))

        if move in self.get_simple_moves(init_loc):
            self.spots[final_loc[0]][final_loc[1]] = self.get_spot_info(init_loc)
            self.spots[init_loc[0]][init_loc[1]] = self.EMPTY_SPOT

            # Here we are promoting a checker piece if it has reached it's king row
            if final_loc in self.get_kings_row_positions(self.P1 if self.player_turn else self.P2):
                self.spots[final_loc[0]][final_loc[1]] = self.P1_K if self.player_turn else self.P2_K

            if switch_player_turn:
                self.player_turn = not self.player_turn

        elif move in self.get_capture_moves(init_loc):
            distx = final_loc[0]-init_loc[0]
            disty = final_loc[1]-init_loc[1]
            enemy_piece_loc = [(distx)//2 + init_loc[0], (disty)//2 + init_loc[1]]
            
            self.spots[final_loc[0]][final_loc[1]] = self.spots[init_loc[0]][init_loc[1]]
            self.spots[init_loc[0]][init_loc[1]] = self.EMPTY_SPOT
            self.spots[enemy_piece_loc[0]][enemy_piece_loc[1]] = self.EMPTY_SPOT

            # Here we are promoting a checker piece if it has reached it's king row
            if final_loc in self.get_kings_row_positions(self.P1 if self.player_turn else self.P2):
                self.spots[final_loc[0]][final_loc[1]] = self.P1_K if self.player_turn else self.P2_K

        else:
            raise ValueError("Not a legal move.")

        if (self.is_game_over()):
            self.Current_Winner = current_player
            self.game_over = True

    def get_potential_spots_from_moves(self, moves):
        """
        Get's the potential spots for the board if it makes any of the given moves.
        If moves is None then returns it's own current spots.
        """
        if moves is None:
            return self.spots
        potential_spots = []
        for move in moves:
            newBoard = Board(self.spots)
            newBoard.make_move(move)
            potential_spots.append(newBoard.spots) 
        return potential_spots
        
    def insert_pieces(self, pieces_info):
        """
        Inserts a set of pieces onto a board.

        pieces_info is in the form: [[x1, y1, piece1], [x2, y2, piece2], ..., [xn, yn, piecen]]
        """
        for piece_info in pieces_info:          
            self.spots[piece_info[0]][piece_info[1]] = piece_info[2]
        
    def get_symbol(self, location):
        """
        Gets the symbol for what should be at a board location.
        """
        if self.spots[location[0]][location[1]] == self.EMPTY_SPOT:
            return " "
        elif self.spots[location[0]][location[1]] == self.P1:
            return "o"
        elif self.spots[location[0]][location[1]] == self.P2:
            return "x"
        elif self.spots[location[0]][location[1]] == self.P1_K:
            return "O"
        elif self.spots[location[0]][location[1]] == self.P2_K:
            return "X"
        else:
            raise ValueError("Not a valid player number. Availible player numbers are: " +str(self.P1)+", "+str(self.P2))
    
    def print_board(self):
        """
        Prints a string representation of the current game board.
        """
        norm_line = "|-----|-----|-----|-----|-----|-----|-----|-----|"
        print(norm_line)
        for j in range(self.HEIGHT):
            row = "|  "
            for i in range(self.WIDTH):
                row = row + self.get_symbol([i, j]) + "  |  "
            print(row)
            print(norm_line)            


from checker_players.Checker_Environment.Board import Board
class Player:

    EMPTY_SPOT = Board.EMPTY_SPOT
    P1 = Board.P1
    P2 = Board.P2
    P1_K = Board.P1_K
    P2_K = Board.P2_K
    BACKWARDS_PLAYER = Board.BACKWARDS_PLAYER
    HEIGHT = Board.HEIGHT
    WIDTH = Board.WIDTH

    def __init__(self, player_number):
        self.player_number = player_number
        
    def play(self, board):
        pass

    def learn(self, reward, board):
        pass


from checker_players.Runners.Loggers.Logger import Logger
import time
import copy

'''
This simple Logger notes who won the game, how many turns the game took to terminate, 
the board state at each turn, and how much time each each turn took. If the "winner of the game"
is set to None then that implies that the game did not terminate. 
'''

class SimpleLogger(Logger):
    winner = None
    current_turn_number = 0
    board_state_history = []
    board_time_per_move_history = []

    time_at_last_move = None

    def __init__(self):
        pass

    def begin(self):
        self.time_at_last_move = time.process_time()

    def apply(self, board, player1, player2):
        self.current_turn_number += 1

        self.board_state_history.append(copy.deepcopy(board.spots))

        self.board_time_per_move_history.append(time.process_time() - self.time_at_last_move)
        self.time_at_last_move = time.process_time()

    def end(self, board, player1, player2):
        self.winner = board.Current_Winner
        pass

    def exit(self, board, player1, player2):
        return self.winner, self.current_turn_number, self.board_state_history, self.board_time_per_move_history


import numpy as np

class TrainRunner:

    def __init__(self, board, player1, player2, logger):
        self.board = board
        self.player1 = player1
        self.player2 = player2
        self.logger = logger

    def learn_games_for_player1(self, num_of_games, turn_cutoff = 100):
        for num_game in range(num_of_games):
            self.board.reset_board()
            self.logger.begin()
            board_states = []
            winner = None
            for _ in range(turn_cutoff):
                if self.board.is_game_over:
                    winner = self.board.Current_Winner
                    break
                if self.board.player_turn:
                    board_states.append(self.board.spots)
                curr_player = self.player1 if self.board.player_turn else self.player2
                self.board.make_move(curr_player.play(self.board))
                self.logger.apply(self.board, self.player1, self.player2)
            self.logger.end()
            self.player1.learn(winner, board_states)
        return self.logger.exit()

    def learn_games(self, num_of_games, turn_cutoff = 100):
        for num_game in range(num_of_games):
            if num_game % 100 == 0:
                print(num_game)
            self.board.reset_board()
            self.logger.begin()
            board_states_p1 = []
            board_states_p2 = []
            winner = None
            for _ in range(turn_cutoff):
                if self.board.game_over:
                    winner = self.board.Current_Winner
                    break
                if self.board.player_turn:
                    board_states_p1.append(self.board.spots)
                else:
                    board_states_p2.append(self.board.spots)
                curr_player = self.player1 if self.board.player_turn else self.player2
                self.board.make_move(curr_player.play(self.board))
                self.logger.apply(self.board, self.player1, self.player2)
            self.logger.end(self.board, self.player1, self.player2)
            self.player1.learn(winner, board_states_p1)
            self.player2.learn(winner, board_states_p2)
        return self.logger.exit(self.board, self.player1, self.player2)


import torch
import torch.nn as nn
from checker_players.Players.Player import Player
from checker_players.Checker_Environment.Board import Board
from checker_players.Interpretters.FlatFeaturesInterpretter import FlatFeaturesInterpretter

class CheckersNet(nn.Module):
    def __init__(self):
        super().__init__()
        self.fc1 = nn.Linear(32, 128)
        self.fc2 = nn.Linear(128, 128)
        self.fc3 = nn.Linear(128, 128)
        self.fc4 = nn.Linear(128, 128)
        self.fc5 = nn.Linear(128, 64)
        self.fc6 = nn.Linear(64, 32)
        self.fc7 = nn.Linear(32, 1)

    def forward(self, x):
        x = torch.relu(self.fc1(x))
        x = torch.relu(self.fc2(x))
        x = torch.relu(self.fc3(x))
        x = torch.relu(self.fc4(x))
        x = torch.relu(self.fc5(x))
        x = torch.relu(self.fc6(x))
        return self.fc7(x)

class MultilayerPerceptron(Player):
    mlpEvaluator = CheckersNet()
    optimizer = torch.optim.SGD(mlpEvaluator.parameters(), lr=1e-3, momentum=0.9)
    interpretter = FlatFeaturesInterpretter()
    
    def __init__(self, player_number):
        self.player_number = player_number

    def board_state_to_tensor(self, board_state):
        board_state = self.interpretter.interpret(board_state)
        if self.player_number == self.P1:
            board_state = board_state
        else:
            board_state = [-1*x for x in board_state];
        return torch.tensor([board_state], dtype = torch.float32)

        
    def play(self, board):
        best_move = None
        best_evaluation = float("-inf")
        for move in board.get_possible_next_moves():
            new_board = Board(board.spots, P1_turn= board.player_turn)
            curr_evaluation = self.mlpEvaluator(self.board_state_to_tensor(new_board.spots))[0]
            best_move = move if curr_evaluation > best_evaluation else best_move
            best_evaluation = curr_evaluation if curr_evaluation > best_evaluation else best_evaluation
        return best_move

    def learn(self, winner, board_states):
        for i, state in enumerate(board_states):
            self.mlpEvaluator.train()
            self.optimizer.zero_grad()
            self.mlpEvaluator(self.board_state_to_tensor(state))
            loss = None
            if (winner == self.player_number):
                loss = torch.tensor([-1*i/len(board_states)], requires_grad=True)
            else:
                loss = torch.tensor([i/len(board_states)], requires_grad=True)
            loss.backward()
            self.optimizer.step()



class Logger:
    def __init__(self):
        pass

    def begin(self, **kwargs):
        pass

    def apply(self, board, player1, player2):
        pass

    def end(self, board, player1, player2):
        pass

    def exit(self, board, player1, player2):
        pass


from checker_players.Players.Player import Player
from checker_players.Interpretters.DoubleColumnInterpretter import DoubleColumnInterpretter
import random 
class RandomPlayer(Player):

    def play(self, board):
        return random.choice(board.get_possible_next_moves())

    # The random player does no learning
    def learn(reward, board):
        pass


class TestRunner:

    def __init__(self, board, player1, player2, logger):
        self.board = board
        self.player1 = player1
        self.player2 = player2
        self.logger = logger

    def play_single_game(self, turn_cutoff = 100):
        self.logger.begin()
        for _ in range(turn_cutoff):
            if self.board.game_over:
                break
            curr_player = self.player1 if self.board.player_turn else self.player2
            self.board.make_move(curr_player.play(self.board))
            self.logger.apply(self.board, self.player1, self.player2)
        self.logger.end(self.board, self.player1, self.player2)
        return self.logger.exit(self.board, self.player1, self.player2)

    def play_games(self, num_of_games, turn_cutoff = 100):
        for num_game in range(num_of_games):
            self.board.reset()
            self.logger.begin()
            for _ in range(turn_cutoff):
                if self.board.game_over:
                    break
                curr_player = self.player1 if self.board.player_turn else self.player2
                self.board.make_move(curr_player.play(self.board))
                self.logger.apply(self.board, self.player1, self.player2)
            self.logger.end(self.board, self.player1, self.player2)
        return self.logger.exit(self.board, self.player1, self.player2)



from checker_players.Runners.Loggers.Logger import Logger
from checker_players.Runners.TestRunner import TestRunner 
from checker_players.Players.RandomPlayer import RandomPlayer
from checker_players.Checker_Environment.Board import Board
from checker_players.Runners.Loggers.SimpleLogger import SimpleLogger 

def evaluate_against_random_player(player1, player2):
    win_rateP1 = 0
    num_games = 100
    for i in range(num_games):
        board = Board()
        logger = Logger()
        random = RandomPlayer(board.P2)
        runner = TestRunner(board, player1, random, logger)
        runner.play_single_game(1000)
        winner = board.Current_Winner
        win_rateP1 = win_rateP1 + 1/num_games if winner == board.P1 else win_rateP1

    win_rateP2 = 0
    for i in range(num_games):
        board = Board()
        logger = Logger()
        random = RandomPlayer(board.P1)
        runner = TestRunner(board, random, player2, logger)
        runner.play_single_game(1000)
        winner = board.Current_Winner
        win_rateP2 = win_rateP2 + 1/num_games if winner == board.P2 else win_rateP2
    
    return win_rateP1, win_rateP2



class TrainLogger(Logger):
    winner_history = []
    current_game_number = 0
    player1WRHistory = []
    player2WRHistory = []


    def __init__(self):
        pass

    def begin(self):
        pass

    def apply(self, board, player1, player2):
        if board.is_game_over():
            self.winner_history.append(board.P1 if (board.player_turn) and board.get_possible_next_moves() else board.P2)

    def end(self, board, player1, player2):
        self.current_game_number += 1
        if self.current_game_number % 100 == 0:
            player1WR, player2WR = evaluate_against_random_player(player1, player2)
            self.player1WRHistory.append(player1WR)
            self.player2WRHistory.append(player2WR)


    def exit(self, board, player1, player2):
        return self.current_game_number, self.winner_history, self.player1WRHistory, self.player2WRHistory


from checker_players.Players.Player import Player
from checker_players.Interpretters.DoubleColumnInterpretter import DoubleColumnInterpretter
from checker_players.Checker_Environment.Board import Board
from checker_players.Interpretters.DoubleColumnInterpretter import DoubleColumnInterpretter

#from tensorflow import keras
#from tensorflow.keras import layers as kl

import torch
import torch.nn as nn

import numpy as np

class DQNPlayer(Player):
    def __init__(self, player_number, model=0): #I think we want a shallow copy of model
        super().__init__(player_number)
        
        self.model = model
        if not self.model:
            self.model = self.create_model()
        
        
    def get_best_move(self, board):
        possible_moves = board.get_possible_next_moves()
        q_values = np.array((possible_moves))
        for i, move in enumerate(possible_moves):
            q_values[i] = self.model(move)
        return possible_moves(np.argmax(q_values))
        
        
    def play(self, board):
        return self.get_best_move(board)
    
    
    def learn(self, logger_exit, reward, learning_rate = 0.01):
        winner, current_turn_number, board_state_history, _ = logger_exit
        for turn, board in enumerate(board_state_history):
            y_pred = self.model(board)
            loss = reward[turn-1]
            self.model.zero_grad()
            loss.backward()
            
            with torch.no_grad():
                for param in self.model.parameters():
                    param -= learning_rate * param.grad
        
    
    def create_model(self):
        model = torch.nn.Sequential(
            [
                #we can make the input 32, since half of all pieces are empty
                nn.Flatten(input_shape=(8, 8)),
                nn.Linear(64, 32), nn.ReLU(),
                nn.Linear(32, 16), nn.ReLU(),
                nn.Linear(16,  8), nn.ReLU(),
                nn.Linear( 8,  4), nn.ReLU(),
                nn.Linear( 4,  1)
            ]
        )
        return model
        
        
        


from checker_players.Players.MinimaxPlayer import MinimaxPlayer
from checker_players.Interpretters.DoubleColumnInterpretter import DoubleColumnInterpretter
from checker_players.Checker_Environment.Board import Board

class AlphaBeta(MinimaxPlayer):    
    def get_best_board_evaluation(self, board, depth_cutoff, alpha, beta, max_step = True):
        if depth_cutoff == 0:
            return self.evaluate_board(board)

        possible_moves = board.get_possible_next_moves()

        if len(possible_moves) == 0:
            return self.evaluate_board(board)

        best_value = None
        for move in possible_moves:
            new_board = Board(board.spots, P1_turn= board.player_turn)
            new_board.make_move(move)
            value = self.get_best_board_evaluation(new_board, depth_cutoff-1, alpha, beta, max_step = max_step if (new_board.player_turn == board.player_turn) else not max_step)
            best_value = value if not best_value else best_value
            if max_step:
                best_value = max(best_value, value)
                alpha = max(alpha, value)
            else:
                best_value = min(best_value, value)
                beta = min(beta, value)

            if alpha >= beta:
                break
        
        return best_value

    def get_best_move(self, board, depth_cutoff):
        '''
        1. get all the possible moves
        2. for each move, pretend we did that move.
        3. use get_board_evaluation to see how good the move was
        4. choose the best move
        '''
        best_move = None 
        best_evaluation = float('-inf')
        alpha = float("-inf")
        beta = float("inf")

        possible_moves = board.get_possible_next_moves()

        for move in possible_moves:
            if not best_move:
                best_move = move
            new_board = Board(board.spots, P1_turn= board.player_turn)
            new_board.make_move(move)
            curr_evaluation = self.get_best_board_evaluation(new_board, depth_cutoff-1, alpha, beta, max_step = (new_board.player_turn == board.player_turn))
            best_move = move if curr_evaluation > best_evaluation else best_move
            best_evaluation = curr_evaluation if curr_evaluation > best_evaluation else best_evaluation

        return best_move

    def play(self, board):
        return self.get_best_move(board, self.depth_cutoff)

    # The Minimax player does no learning
    def learn(self, reward, board):
        pass


from checker_players.Interpretters.Interpretter import Interpretter

class DoubleColumnInterpretter(Interpretter):
    def interpret(self, board):
        P1_column = [[x,y] for x in range(board.WIDTH) for y in range(board.HEIGHT) if board.get_spot_info([x,y]) in (board.P1, board.P1_K)]
        P2_column = [[x,y] for x in range(board.WIDTH) for y in range(board.HEIGHT) if board.get_spot_info([x,y]) in (board.P2, board.P2_K)]
        return {board.P1: P1_column, board.P2: P2_column}


from checker_players.Interpretters.Interpretter import Interpretter

class FlatFeaturesInterpretter(Interpretter):
    def interpret(self, board_state):
        features = []
        for x, row in enumerate(board_state):
            for y, element in enumerate(row):
                if (x+y) % 2 == 1:
                    features.append(board_state[x][y])
        return features